#include <Wire.h>
#include "Adafruit_BME680.h"
#include <Adafruit_NeoPixel.h>
#include <Ticker.h>
#include <Encoder.h>
#include <Adafruit_GFX.h>
#include <Adafruit_IS31FL3731.h>



//--------- Global definitions
// Encoder-Library http://www.pjrc.com/teensy/td_libs_Encoder.html
Encoder button_encoder(14,12); // Button rotation encoder object
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(2,13,NEO_GRBW + NEO_KHZ800);

#define DELAY_MYS_MAX 16383
#define NOT_A_GENUINE_RESULT 12917167
#define COOLDOWN false
#define HEATUP true

Adafruit_BME680 bme; // I2C

/* |StrControl| contains the measurement configuration
 *  ("control string") which is received from the AI system
 */
String StrControl ="";

/* Those flags are changed by interrupts. */
volatile int hasControlString = 0; /* 0: no control string received yet, |StrControl| empty
                                    * 1: control string received
                                    * 2: there is new control string that replaces the old one!
                                    *    Discard the current measurement with the old control string,
                                    *    then set |hasControlSting = 1|.
                                     */
volatile int buttonPressedFlag = 0;
volatile int waitingForStart = 1;
uint16_t cool_temp;

int buttonOffset = 0; /* contains the offset in the button rotation. Is explained further below */

enum PixelDir {RIGHT, LEFT};
enum PixelWhen {PREVIOUS, CURRENT};

enum StringType {CONTROL, PHYSVAL} stringType;
String serialRead(bool = false, bool = false);
void cooldown(bool, uint16_t = 0);


//--------- CharlieWing Matrix Display Definitions
Ticker matrixOutput;
volatile int matrixOutput_index = 0; /* current position in matrix */
String matrixOutput_text  = " ";     /* output string as global variable */
volatile int matrixOutput_wait  = 0; /* to wait until the (scrolling) text has run through */

Adafruit_IS31FL3731_Wing matrix = Adafruit_IS31FL3731_Wing();

/* matrixUpdate: updates the matrix output */
void matrixUpdate(){
  int numberPixels=(matrixOutput_text.length())*6;
  if (numberPixels > 15) { /* scrolling */
    if (matrixOutput_index >= -numberPixels) {
      matrix.clear();
      matrix.setCursor(matrixOutput_index,0);
      matrix.print(matrixOutput_text);
      matrixOutput_index--;
    } else {
      matrixOutput_index = 0;
      matrixOutput_wait=0;
    }
  } else { /* only three symbols max, no scrolling */
    matrix.setCursor(0,0);
    matrix.print(matrixOutput_text);
    matrixOutput_wait  = 0; /* no waiting */
  }
}

/* matrixDisplay: sets the text to output */
void matrixDisplay(String text, int wait) {
  if (matrixOutput_text  != text) { /* update only when text is different from before */
    matrix.clear();
    matrixOutput_index = 0;
    matrixOutput_text  = text;
    matrixOutput_wait  = wait;
    while (matrixOutput_wait) delay(10); /* wait until text is completely displayed once */
  } 
};

//--------- NeoPixel Definitions
struct PixelState{       /* stores the current and the previous NeoPixel state for both pixels */
  int red;
  int green;
  int blue;
} pixelState[2][2];

/* setPixelColor: sets the specified Pixel to its color and saves the previous state */
void setPixelColor(enum PixelDir pixelDir, int red, int green, int blue){
   struct PixelState now = {red, green, blue};
  
   pixels.setPixelColor(pixelDir, red, green, blue);
   pixels.show();

   memcpy(&pixelState[pixelDir][PREVIOUS], &pixelState[pixelDir][CURRENT], sizeof(pixelState[pixelDir][CURRENT]));
   memcpy(&pixelState[pixelDir][CURRENT], &now, sizeof(now));
}

/* setPreviousPixelColor: restores the previous pixel color for the specified pixel and sets it as the current one */
void setPreviousPixelColor(enum PixelDir pixelDir){
   struct PixelState flip;

   pixels.setPixelColor(pixelDir, pixelState[pixelDir][PREVIOUS].red,
                                  pixelState[pixelDir][PREVIOUS].green,
                                  pixelState[pixelDir][PREVIOUS].blue);
   pixels.show();
   
   memcpy(&flip, &pixelState[pixelDir][CURRENT], sizeof(pixelState[pixelDir][CURRENT]));
   memcpy(&pixelState[pixelDir][CURRENT], &pixelState[pixelDir][PREVIOUS], sizeof(pixelState[pixelDir][PREVIOUS]));
   memcpy(&pixelState[pixelDir][PREVIOUS], &flip, sizeof(flip));
}


//--------- button definitions
/* buttonPress_ISR: Interrupt service routine executed on button press */
void buttonPress_ISR() {
   int i;
  /* button debouncing */
   int old_key = digitalRead(2);
   int new_key;
   delayMicroseconds(10000); 
   new_key = digitalRead(2);
   if (new_key != old_key){
      return;
   }

   /* indicate that the button was pressed to start the board operation */
   if (waitingForStart){
      waitingForStart = 0;
      return;
   }

  for (i = 0; i < 25; i++)
    delayMicroseconds(DELAY_MYS_MAX);

  if (digitalRead(2)==LOW) { /* button was held down to determine the selected drink class */

      matrixDisplay(String("..."),0); 
      serialRead();
      Serial.println(String("CHECK:") + String(buttonRead()).c_str());
  }
  else
    buttonPressedFlag = 1;
  }

/* buttonRead: reads the rotation of the button
 * substracts the offset to make the top position == 0
 */
int buttonRead(){
  return button_encoder.read() - buttonOffset;
  }



//--------- SETUP()
void setup(){


  Serial.begin(115200, SERIAL_8E1);

  pixels.begin(); /* initlalize NeoPixels */
  delay(1);

  setPixelColor(RIGHT,0,0,0); /* turn NeoPixels off */
  setPixelColor(LEFT,0,0,0);

  Wire.begin(); /* initlailze I2C bus */
  if (Wire.status() != I2C_OK)
	  Serial.println("O:Something wrong with I2C");

  matrix.begin(); /* initialize CharlieWing Display Matrix */
  delay(10);
  matrix.clear(); 
  matrix.setTextColor(60); /* limit brightness */ 
  matrixOutput.attach(0.08, matrixUpdate); /* cyclic update */

  /* button interrupt configuration */
  attachInterrupt(2, buttonPress_ISR, FALLING);

  /* blink right NeoPixel while waiting for start */
  while (waitingForStart) {
  setPixelColor(RIGHT,0,0,40); // blue

  delay(200);
  setPixelColor(RIGHT,0,0,0); // off

  delay(200);
  }

  /* aftr receiving the start button push, detect the button position to determine the correct offset */
  buttonOffset = button_encoder.read();

 
  /* Gas sensor check */
  if (!bme.begin(118)) {
   Serial.println("O:Could not find a valid BME680 sensor, check wiring!");
    do {} while(1);
  }


  setPixelColor(RIGHT,40,40,0); // yellow

  /* display welcome message */
  matrixDisplay(String("Welcome"),1);  
  matrixDisplay(String(""),1);
  
  /* signaling the AI system the need for a control string*/
  Serial.println("ID");
   /* wait for control string */
  while (!hasControlString){
    delay(500);
    serialRead();
  }

  /* thr first temperature of the control string is marked the cool_temp.*/
  cool_temp = String(StrControl.substring(StrControl.indexOf("TEMP:")+5, StrControl.indexOf(",",StrControl.indexOf("TEMP:")))).toInt();

  setPixelColor(RIGHT,00,30,0); // green

}


//--------- SERIAL INPUT FUNCTION

String serialRead(bool control /* = false */, bool forced /* = false */){
  int received = 0;
  String decision;
  String inputString = "";

  while (forced && !Serial.available())
    delay(100);
  
  while (Serial.available()) {
    received = 1;
    // get the new byte:
    char inChar = (char)Serial.read();
    // add it to the inputString:
    inputString += inChar;
    
  }
  if (received) {
    decision = inputString.substring(0, 1);
    if (decision.equals("T") || control == true)
        stringType = CONTROL;
    else
        stringType = PHYSVAL;  
  
    if (stringType == CONTROL)
    {
		if (inputString.indexOf(String(";;")) != -1){
		  if (control == true)
			return inputString;
		}
		else{
		  delay(200);
		inputString += serialRead(true, true);
		}

		StrControl = inputString;
		
		if  (hasControlString == 0) /* no control string arrived yet */
			 hasControlString =  1; /* our first control string */
		else hasControlString =  2; /* update: another control string! */ 
    } else {
       matrixDisplay(String(inputString) + " " + String(inputString),1);  
       matrixDisplay(String(""),1);
    }
  }
  return inputString;
      
}

void setGasHeater(uint16_t temperature, uint16_t heatingDuration){
  bme.setGasHeater(temperature, heatingDuration);
  String out = String("O:temp:") + String(temperature) + String(";hd:") + String(heatingDuration);
  Serial.println(out.c_str());
  
  }


//--------- PERFORM MEASUREMENT FUNCTION
/* performMeasurement: performs a measurement series and returns the string containing the results
 * if |buttonSensitive == 1| then it will abort the measurement when the button is pressed
 * and return an empty string to indicate the abortion.
 */
String performMeasurement(int buttonSensitive){
  float hum = 0, temp = 0, press = 0;
  int lastIndex  = -1, nextIndex=0;
  uint16_t heatingDurationDefault = 75;
  uint16_t temperatureDefault  = 200;
  uint16_t temperature;
  uint16_t heatingDuration;
  int buttonDirection;
  String strControlTemp, strControlHeatingDur, strTemp, strVal;
  int cycleRunning = 1;
  String result = "";

  Serial.println("O:performingMeasurement...");

  do {
  /* read surrounding information, currently not done. */
  // temp = bme.readTemperature();
  // hum  = bme.readHumidity();
  // press = bme.readPressure();
  temp = 0;
  hum = 0;
  press = 0;
  /* read current button rotation to determine training class */
  buttonDirection = buttonRead();

  /* if |!(hasControlString == 1)| at the end of the measurement, the control string has been changed in between */
  hasControlString = 1;
  
  /* extract temperature profile and heating duration from control string
   * and set it. if no heating duration is found, set to default.
   */
  strControlTemp = StrControl.substring(StrControl.indexOf("TEMP:")+5, StrControl.indexOf(";",StrControl.indexOf("TEMP:")));
  strControlHeatingDur = StrControl.substring(StrControl.indexOf("HTDR:")+5);
  heatingDuration = strControlHeatingDur.toInt();
  if (heatingDuration == 0) heatingDuration = heatingDurationDefault;

  while (cycleRunning) {
    uint32_t val;
     /* look for next colon */
     nextIndex   = strControlTemp.indexOf(',',lastIndex+1);      
     if (nextIndex <= 0) { /* end */
       nextIndex    = strControlTemp.length(); /* last value */
       cycleRunning = 0; /* end */
     }
     strTemp     = strControlTemp.substring(lastIndex+1,nextIndex);
     temperature = strTemp.toInt();
     if (!temperature) temperature = temperatureDefault;  
     lastIndex = nextIndex;
     cooldown(HEATUP, temperature);
     bme.setGasHeater(temperature,heatingDuration);
     do {
      val = bme.readGas();
     } while (val == 0 || val == NOT_A_GENUINE_RESULT);
    
     strVal = (String(val)+","); 
     result += strVal;
	 
     serialRead();
     if (buttonSensitive & buttonPressedFlag) /* the button has been pushed during measurement */
        return "";
  }
    } /* cycle running */
    while (hasControlString != 1); /* if the control string did change, repeat the procedure */
    return result.substring(0,result.length()-1);
}

//--------- COOLDOWN FUNCTION
/* cooldown:	performs the cooldown phase where the sensor is held on the specified temperature until a saturation
 *		 		is reached (i.e. the sensor reading doesn't change anymore)
 */
void cooldown(bool cooldown, uint16_t temperature /* == 0*/){

    uint32_t val[2];
    int toggle = 0;
    float treshold;
    int minimum = 10;

    if (!temperature)
      setGasHeater(cool_temp, 1000);
    else
      setGasHeater(temperature, 1000);

    while(NOT_A_GENUINE_RESULT == bme.readGas())
      ;
    
    val[0] = bme.readGas();
    val[1] = bme.readGas();
    do {
      
      val[toggle] = bme.readGas();
      treshold = (float)val[toggle] / (float)val[(toggle + 1) % 2];
      ++toggle %= 2;
      if (cooldown == COOLDOWN)
        Serial.println(String("O:cooldown:") + String(treshold, 6));
      else
        Serial.println(String("O:heatup:") + String(treshold, 6));
      if (minimum)
        minimum--;
    
    } while (treshold < 0.998 || treshold > 1.002 || minimum);
}


//--------- LOOP()
void loop() {
  String tempEnv;
  

  static enum states {READY, MEASURING, LOCKED} state = LOCKED;
  
  switch (state){
    case READY:
    
        setPixelColor(LEFT,15,15,15); /* white */
        cooldown(COOLDOWN);
        
        state = MEASURING;
        break;

    case MEASURING:
                                                                                                     
        setPixelColor(LEFT,0,30,0); /* green */
        {
          String substVal = "SUBST:" + performMeasurement(0) + ";";
          serialRead();
		  
          Serial.println(substVal.c_str());
		  
          delay(100);
          while(serialRead().length() == 0) //wait for classification string to arrive
          ;
                                                                                                  
          state = LOCKED;
        }
        break;

    case LOCKED:
        
         
        setPixelColor(LEFT,30,0,30); /* magenta */
        cooldown(COOLDOWN);

        setPixelColor(LEFT,30,0,0); /* red */
        tempEnv = "ENV:" + performMeasurement(0) + ";";

        Serial.println(tempEnv.c_str());
        

        state = READY;
        break;
    }
}



