% initialization
load("allTemperatures30-400.mat");
tempFrom = 30;
tempTo = 400;
classes = unique(Ycategories);
numClasses = numel(classes);
Xarray = table2array(X);
numSamples = size(Xarray,1);
numFeatures = size(Xarray,2);

figure
colormap(zeros(numClasses, 3));
hsv_t = hsv;
colormap(hsv_t)

% Plot the annotation box (heading)
annotation('textbox',[0.13,0.86,0.775,0.12],'String',...
    ['This plot shows all datasets of full measurements over the '...
    'temperature range from ' int2str(tempFrom) ' ' char(176) 'C (Feature 1) to '...
    int2str(tempTo) ' ' char(176) 'C (Feature ' int2str(numFeatures) ') measured in '...
    'steps of 1 ' char(176) 'C. '...
    'Select two features (x-axis) by clicking on the plot, afterwards press ENTER '...
    'for a detail view. Abort: Ctrl+C'], 'FontUnits', 'normalized', 'FontSize', 0.025);

% Plot whole dataset
subplot(11,1,2:6);
hold on
for i = 1:numClasses
    plot(1:numFeatures,Xarray(Ycategories == classes(i),:), 'color', hsv_t(i,:));   
end
xlabel("Feature");
ylabel("Sensor response");

% Plot the legend
legendClasses = [];
for i = 1:numClasses
    legendClasses = [legendClasses, plot(NaN, 'Color', hsv_t(i,:))];
end
legend(legendClasses,cellstr(classes),'Location','northwest')

% Feature view loop
while(true)
% Select a feature for one-dimensional display
[selectedFeature, ~] =  ginput;
selectedFeature = round(selectedFeature);
% feature must be between 1 and numFeatures
selectedFeature = min(selectedFeature,numFeatures);
selectedFeature = max(selectedFeature,1);

    subplot(11,1,7:11);
    for i=1:numClasses
            plot(Xarray((Ycategories == classes(i)),selectedFeature(1)),Xarray((Ycategories == classes(i)),selectedFeature(2)),'*', 'Color', hsv_t(i,:));
            hold on
    end
    hold off
    xlabel(['Feature ' int2str(selectedFeature(1)) ' sensor response']);
    ylabel(['Feature ' int2str(selectedFeature(2)) ' sensor response']);
end